
#include <iostream>

#include <Windows.h>
#include <imagehlp.h>

#include <boost/filesystem.hpp>
#include <boost/filesystem/fstream.hpp>
#include <boost/format.hpp>
#include <boost/spirit/home/support/detail/endian.hpp>
#include <boost/foreach.hpp>

#pragma comment(lib, "imagehlp.lib")

namespace endian = boost::spirit::endian;

class MzHeader : boost::noncopyable {
public:
  char signature[2];
  endian::ulittle8_t unknown[58];
  endian::ulittle32_t peHeaderAddress;
  bool IsValid() const {
    return std::equal(signature, &signature[_countof(signature)], "MZ");
  }
};

class PeHeader : boost::noncopyable {
public:
  char signature[4];
  endian::ulittle8_t unknown1[2];
  endian::ulittle16_t sectionCount;
  endian::ulittle8_t unknown2[4];
  endian::ulittle32_t symbolTableAddress;
  endian::ulittle32_t symbolTableCount;
  endian::ulittle16_t optionalHeaderSize;
  endian::ulittle16_t attributes;
  bool IsValid() const {
    if (!std::equal(signature, &signature[_countof(signature)], "PE\0\0")) {
      return false;
    }
    return true;
  }
};

class DataDirectory {
public:
  endian::ulittle32_t virtualAddress;
  endian::ulittle32_t size;
};

class PeOptionalHeader : boost::noncopyable {
public:
  endian::ulittle16_t signature;
  endian::ulittle8_t unknown[94];
  DataDirectory directories[16];
};

class SectionHeader : boost::noncopyable {
public:
  char name[8];
  endian::ulittle32_t virtualSize;
  endian::ulittle32_t virtualAddress;
  endian::ulittle32_t size;
  endian::ulittle32_t address;
  endian::ulittle32_t relocateAddress;
  endian::ulittle8_t unknown[12];
};

class ImportDescriptor : boost::noncopyable {
public:
  endian::ulittle32_t inThunkVirtualAddress;
  endian::ulittle8_t unknown[8];
  endian::ulittle32_t dllNameVirtualAddress;
  endian::ulittle32_t outThunkVirtualAddress;
};

class ImportThunk : boost::noncopyable {
public:
  endian::ulittle32_t virtualAddress;
};

bool LoadImage(const boost::filesystem::path &path, std::vector<unsigned char> &out) {
  if (!boost::filesystem::exists(path)) {
    return false;
  }
  boost::filesystem::ifstream ifs(path, std::ios::binary);
  if (!ifs.is_open()) {
    return false;
  }

  // MZwb_[̓ǂݍ
  out.resize(sizeof(MzHeader));
  ifs.read(reinterpret_cast<char*>(&out.front()), out.size());
  const MzHeader *mz = reinterpret_cast<const MzHeader *>(&out.front());
  if (!ifs.good() || !mz->IsValid()) {
    return false;
  }

  // PEwb_[̓ǂݍ
  out.resize(mz->peHeaderAddress + sizeof(PeHeader));
  mz = reinterpret_cast<const MzHeader *>(&out.front());
  ifs.seekg(0, std::ios::beg);
  ifs.read(reinterpret_cast<char*>(&out.front()), out.size());
  const PeHeader *pe = reinterpret_cast<const PeHeader *>(&out.front() + mz->peHeaderAddress);
  if (!ifs.good() || !pe->IsValid()) {
    return false;
  }

  // PEIvViwb_[ZNVXg܂ł̓ǂݍ
  out.resize(mz->peHeaderAddress + sizeof(PeHeader) + pe->optionalHeaderSize + sizeof(SectionHeader) * pe->sectionCount);
  mz = reinterpret_cast<const MzHeader *>(&out.front());
  pe = reinterpret_cast<const PeHeader *>(&out.front() + mz->peHeaderAddress);
  ifs.seekg(0, std::ios::beg);
  ifs.read(reinterpret_cast<char*>(&out.front()), out.size());
  if (!ifs.good()) {
    return false;
  }

  // ZNV̓ǂݍ
  const SectionHeader *sectionList = reinterpret_cast<const SectionHeader *>(&out.front() + mz->peHeaderAddress + sizeof(PeHeader) + pe->optionalHeaderSize);
  for (unsigned int i = 0; i < pe->sectionCount; i++) {
    if (sectionList[i].virtualSize == 0) {
      continue;
    }
    if (out.size() < sectionList[i].virtualAddress + sectionList[i].virtualSize) {
      out.resize(sectionList[i].virtualAddress + sectionList[i].virtualSize);
      mz = reinterpret_cast<const MzHeader *>(&out.front());
      pe = reinterpret_cast<const PeHeader *>(&out.front() + mz->peHeaderAddress);
      sectionList = reinterpret_cast<const SectionHeader *>(&out.front() + mz->peHeaderAddress + sizeof(PeHeader) + pe->optionalHeaderSize);
    }
    ifs.seekg(sectionList[i].address, std::ios::beg);
    ifs.read(reinterpret_cast<char*>(&out[sectionList[i].virtualAddress]), (std::min)(sectionList[i].virtualSize, sectionList[i].size));
    if (!ifs.good()) {
      return false;
    }
  }
  return true;
}

bool ExeParse(const boost::filesystem::path &path) {
  if (!boost::filesystem::exists(path)) {
    std::cout << "t@C܂ł\n";
    return false;
  }

  std::vector<unsigned char> image;
  if (!LoadImage(path, image)) {
    std::cout << "t@CJȂAs\`t@Cł͂܂\n";
    return false;
  }
  const unsigned char * const baseAddr = &image.front();


  const MzHeader &mz = *reinterpret_cast<const MzHeader *>(baseAddr);
  if (!mz.IsValid()) {
    return false;
  }

  const PeHeader &pe = *reinterpret_cast<const PeHeader *>(baseAddr + mz.peHeaderAddress);
  if (!pe.IsValid()) {
    return false;
  }

  const PeOptionalHeader &peOptionalHeader = *reinterpret_cast<const PeOptionalHeader *>(baseAddr + mz.peHeaderAddress + sizeof(PeHeader));
  const DataDirectory &dir = peOptionalHeader.directories[1];
  if (dir.size == 0) {
    return false;
  }

  const ImportDescriptor * desc = reinterpret_cast<const ImportDescriptor *>(baseAddr + dir.virtualAddress);
  for (; desc->inThunkVirtualAddress != NULL; desc++) {
    const char * const dllName = reinterpret_cast<const char *>(baseAddr + desc->dllNameVirtualAddress);
    std::cout << boost::format("%s\n") % dllName;
    const ImportThunk *thunk = reinterpret_cast<const ImportThunk *>(baseAddr + desc->inThunkVirtualAddress); 
    for (; thunk->virtualAddress != NULL; thunk++) {
      if ((thunk->virtualAddress & 0x80000000) != 0) {
        std::cout << boost::format("\t@%d\n") % (thunk->virtualAddress - 0x80000000);
        continue;
      }
      const char * const funcName = reinterpret_cast<const char *>(baseAddr + thunk->virtualAddress + 2);
      std::cout << boost::format("\t%s\n") % funcName;
    }
  }
  return true;
}

int main(const unsigned int argc, const char * const * argv) {
  if (argc !=  2) {
    std::wcout << boost::wformat(L"usage: %s path") % boost::filesystem::path(argv[0]).filename().wstring();
    return 1;
  }
  return ExeParse(argv[1]) ? 0 : 1;
}
